/******************************************************************************
 * $Id: hilite.c,v 1.84 2004/09/22 01:14:22 jimb Exp $
 * @file
 * @author	Jim Brooks http://www.jimbrooks.org
 * @brief	Functions for hiliting src code in a text widget.
 ******************************************************************************/

#define DEFINE_HILITE_GLOBALS
#include "common.h"
#include "hilite.h"

#include "hilite_c.h"
#include "hilite_forth.h"
#include "hilite_perl.h"
#include "hilite_python.h"
#include "widgets.h"
#include "statusbar.h"
#include "text.h"

/*****************************************************************************
 * Must be called to initialize hilite functions.
 *****************************************************************************/
void
InitHilite( void )
{
   InitHilitePerl();
   InitHilitePython();
   return;
}

/*****************************************************************************
 * Load a text widget with highlighted source code.
 * @pre		The text widget must be empty and frozen.
 * @param	pWidget
 *          Pointer to a text widget.
 * @param	pBuffer
 *          Pointer to a buffer containing source code text.
 *          Buffer must have a trailing null char.
 * @param	bufferSize
 *          Size of buffer (excluding trailing null char).
 * @param	srcFileKind
 *          One of SRC_FILE_*
 *****************************************************************************/
void
HiliteLoadCode( GtkWidget* pWidget,
                char*      pBuffer,
                int        bufferSize,
                int        srcFileKind )
{
   gboolean	hilited = FALSE;

  /*
   * Load fonts.
   * (Must be done every time, in case user selects different fonts.)
   */
   pFontComment = gdk_font_load( fontNameHiliteComment );
   pFontKeyword = gdk_font_load( fontNameHiliteKeyword );
   pFontType    = gdk_font_load( fontNameHiliteType );
   pFontSymbol  = gdk_font_load( fontNameHiliteSymbol );

  /*
   * Ensure the buffer is terminated by a null char,
   * otherwise the parsing of it would overrun.
   */
g_return_if_fail( pBuffer[bufferSize] == '\0' );

  /*
   * Is buffer tiny or too big?
   * This code is so slow that trying to hilite a source file > 1MB
   * will make the program appear to hang.
   */
   if ( bufferSize < 10  ||  bufferSize > MAX_HILITE_BUFFER_LEN )
   {
      gtk_text_insert( GTK_TEXT(pWidget),
                       NULL,
                       NULL,
                       NULL,
                       pBuffer,
                       bufferSize );

      if ( bufferSize > MAX_HILITE_BUFFER_LEN )
         PrintStatusbar( BUFFER_TOO_LARGE_TO_HILITE_STR );

      return;
   }

  /*
   * Switch to a language-specific function that
   * will load a text widget with hilited src code.
   */
   switch ( srcFileKind )
   {
      case SRC_FILE_C:
      case SRC_FILE_CPLUSPLUS:
      case SRC_FILE_JAVA:
         hilited = HiliteC( pWidget, pBuffer, bufferSize, srcFileKind );
         break;

      case SRC_FILE_FORTH:
         hilited = HiliteForth( pWidget, pBuffer, bufferSize, srcFileKind );
         break;

      case SRC_FILE_PERL:
         hilited = HilitePerl( pWidget, pBuffer, bufferSize, srcFileKind );
         break;

      case SRC_FILE_PYTHON:
         hilited = HilitePython( pWidget, pBuffer, bufferSize, srcFileKind );
         break;

      default:
         hilited = TRUE; /* just lie -- Bill Clinton */
         gtk_text_insert( GTK_TEXT(pWidget),
                          NULL,
                          NULL,
                          NULL,
                          pBuffer,
                          bufferSize );
         break;
   }

   if ( !hilited )
   {
     /*
      * Hiliting failed.
      * Plan B: insert text without hiliting.
      */
      ClearTextWidget();
      gtk_text_insert( GTK_TEXT(pWidget),
                       NULL,
                       NULL,
                       NULL,
                       pBuffer,
                       bufferSize );
   }

   return;
}

/*****************************************************************************
 * Parse the comment that is being pointed to.
 * Insert the comment as hilited text into a text widget.
 * @param	pWidget
 *          Pointer to a text widget.
 * @param	pCommentStartApprox
 *          This can point to either the exact start of a comment (to "/"),
 *          or can point to whitespace that precedes the comment.
 * @param	pCommentEndDef
 *          Pointer to a string that defines what the end of comment is.
 *          Eg, "/""*" or "\n" for C or C++, resp.
 * @param	bufferSize
 *          Size of buffer, measurement based on "pCommentStart" as the start.
 * @param	pFontComment (GLOBAL)
 *          Hilite font (can be NULL).
 * @param	colorComment (GLOBAL)
 *          Hilite color.
 * @return	If a comment end was present, returns a pointer after it,
 *          or returns NULL.
 *****************************************************************************/
const char*
HiliteComment( GtkWidget*  pWidget,
               const char* pCommentStartApprox,
               const char* pCommentEndDef,
               int         bufferSize )
{
   char*	pCommentEnd;
   int      commentLen;

#ifdef HYPERSRC_PEDANTIC
g_return_val_if_fail( pCommentStartApprox  &&  pCommentStartApprox[0]  &&  bufferSize > 0, NULL );
#endif

  /*
   * Try to find end of comment.
   */
   pCommentEnd = strstr( pCommentStartApprox, pCommentEndDef );

  /*
   * Compute length of comment.
   * If the comment epilogue is missing then use end-of-buffer.
   */
   if ( pCommentEnd != NULL )
      commentLen = (pCommentEnd + strlen(pCommentEndDef)) - pCommentStartApprox;
   else
      commentLen = bufferSize;

  /*
   * Insert comment into text widget.
   */
   gtk_text_insert( GTK_TEXT(pWidget),
                    pFontComment,
                    &colorComment,
                    NULL,
                    pCommentStartApprox,
                    commentLen );

  /*
   * Return pointer to after end of comment, or NULL.
   */
   if ( pCommentEnd )
      return pCommentEnd + strlen(pCommentEndDef);
   else
      return NULL;
}

/*****************************************************************************
 * Hilite a double-quoted string.
 * @param	pWidget
 *          Pointer to a text widget.
 * @param	pFont
 *          Hilite font (NULL can be passed).
 * @param	pColor
 *          Hilite color (NULL can be passed).
 * @param	pStart
 *          Actual start of string (typically, contains leading whitespace).
 * @param	pQuote1
 *          Pointer to first dquote.
 * @param	pBufEnd
 *          Pointer to end of buffer.
 * @param	pAfter (OUT)
 *          Points to next char after the ending dquote.
 * @return	FUNC_OK if successful.
 *****************************************************************************/
FUNC_X
HiliteQuotedString( char         quoteChar,
                    GtkWidget*   pWidget,
                    GdkColor*    pColor,
                    GdkFont*     pFont,
                    const char*  pStart,
                    const char*  pQuote1,
                    const char*  pBufEnd,
                    const char** ppAfter /*OUT*/ )
{
   const char* pc;
   const char* pQuote2;

#ifdef HYPERSRC_PEDANTIC
   if (    pStart == NULL
        || pQuote1 == NULL
        || pBufEnd == NULL
        || pQuote1 < pStart
        || pQuote1 >= pBufEnd )
   {
      WarningMalfunction();
      return FUNC_FAIL;
   }
#endif

  /*
   * Ensure first dquote is pointed to.
   */
   if ( *pQuote1 != quoteChar ) return FUNC_FAIL;

  /*
   * Find second dquote.  Ignore \"
   * There are some C source files out there with very long strings
   * so the length of hilited strings isn't limited.
   */
   pQuote2	= NULL;
   for ( pc = pQuote1+1; pc < pBufEnd; ++pc )		/* skip first dquote */
   {
      if ( *pc == quoteChar )
      {
         pQuote2 = pc;
         break;
      }
      else if ( *pc == '\\' ) /* handles "\"" and "\\" */
      {
         ++pc;     /* skip the following escaped char */
         continue;
      }
   }
   if ( pQuote2 == NULL ) return FUNC_FAIL;	/* was second dquote found? */

  /*
   * Insert dquoted string and any leading whitespace.
   */
   gtk_text_insert( GTK_TEXT(pWidget),
                    pFont,
                    pColor,
                    NULL,
                    pStart,
                    pQuote2 - pStart + 1 );

   *ppAfter = pQuote2 + 1;
   return FUNC_OK;
}

/*****************************************************************************
 * Search for an exact word in a string of words (the entire word must match).
 * Each word in the string must be separated by a leading space,
 * to avoid matches on part of a word.
 * @param	pWords
 *          String of words (each word must be preceded by a space).
 * @param	pSought, soughtLen
 *          Word that is sought (matched against words in pWords).
 *****************************************************************************/
int
FindWord( const char* pWords,
          char*       pSought,
          int         soughtLen )
{
   char*		pMatch;
   const char*	pWordsNext;
   int			i;

#ifdef HYPERSRC_PEDANTIC
   if ( pWords[0] != ' ' )
   {
      DebugMsg( "was passed an incorrect string of words." );
      return FALSE;
   }
#endif

   pWordsNext = pWords;

   for ( i = 0; i < 100; i++ ) /* prevent infinite loop */
   {
      pMatch = strstr( pWordsNext, pSought );

      if ( pMatch == NULL )
         return FALSE;

     /*
      * If the word matches exactly, there should be a leading space -and-
      * either a trailing space or null char.
      */
      if (    *(pMatch-1) == ' '
           && ( pMatch[soughtLen] == ' '  ||  pMatch[soughtLen] == 0 ) )
         return TRUE;

     /*
      * Move past the impartial match.
      */
      pWordsNext = pMatch + soughtLen;

#ifdef HYPERSRC_PEDANTIC
      if ( pWordsNext > (pWords + strlen(pWords)) )
      {
         DebugMsg( "pWordsNext > (pWords + strlen(pWords))" );
         return FALSE;
      }
#endif
   }

   DebugMsg( "Stopping infinite loop in FindString()." );
   return FALSE;
}
