/******************************************************************************
 * $Id: hilite_perl.c,v 1.25 2004/09/22 01:14:22 jimb Exp $
 * @file
 * @author	Jim Brooks http://www.jimbrooks.org
 * @brief	Functions for hiliting Perl code in a text widget.
 ******************************************************************************/

#include "common.h"
#include "hilite_perl.h"

#include "hilite.h"
#include "widgets.h"
#include "statusbar.h"
#include "text.h"

#define INCLUDE_VAR_CHARS	TRUE
#define EXCLUDE_VAR_CHARS	FALSE

/*
 * Perl keywords (well, most of the major/common ones)
 */
static const char keywords_perl[] =
   " if"
   " else"
   " elsif"
   " unless"
   " foreach"
   " for"
   " next"
   " last"
   " redo"
   " until"
   " while"
   " my"
   " return"
   " die"
   " exit"
   " warn"
   " sub"
/* " goto"  is handled/colored specially */
   " eq"
   " ne"
   " lt"
   " gt"
   " cmp"
   " use"
   " require"
   " package"
   " print"
   " push"
   " pop"
   " shift"
   " eval"
   " exists"
   " defined"
   " undef"
   " delete"
   " chop"
   " chomp"
   "                 "; /* padding just in case */

/*
 * Perl symbol chars (excluding "$@%").
 */
const char symbols_perl_other[] =
   " ="
   " =="
   " !="
   " <"
   " >"
   " <="
   " >="
   " +"
   " -"
   " *"
   " /"
   " ||"
   " &&"
   " ?"
   " !"
   " &"
   " |"
   " ^"
   " ~"
   " ,"
   " ["
   " ]"
   " ."
   " <<"
   " >>"
   " ++"
   " --"
   " +="
   " -="
   " *="
   " /="
   " &="
   " ^="
   " |="
   " <<="
   " >>="
   " :"
   "                 "; /* padding just in case */

/*
 * This table includes most but not all symbol chars.
 * Some are excluded because they're specially hilited (eg quote chars).
 */
static char tablePerlSymbolMost[256];

/*****************************************************************************
 * Must be called to initialize utility functions.
 *****************************************************************************/
void
InitHilitePerl( void )
{
   int i;

  /*
   * This table includes most but not all symbol chars.
   * Some are excluded because they're specially hilited (eg quote chars).
   */
   CLEAR( tablePerlSymbolMost, sizeof(tablePerlSymbolMost) );
   for ( i = 0; i < 256; ++i ) tablePerlSymbolMost[i] = IsPerlSymbolChar( (char)i );
   tablePerlSymbolMost[ (int)'"'  ] = 0;
   tablePerlSymbolMost[ (int)'\'' ] = 0;
   tablePerlSymbolMost[ (int)'`'  ] = 0;
   tablePerlSymbolMost[ (int)'['  ] = 0;
   tablePerlSymbolMost[ (int)'#'  ] = 0;

   return;
}

/*****************************************************************************
 * Load a text widget with hilited Perl source code.
 * @pre		The text widget must be empty and frozen.
 * @param	pWidget
 *          Pointer to a text widget.
 * @param   pBuffer
 *			Pointer to a buffer containing source code text.
 *			Buffer must have a trailing null char.
 * @param	bufferSize
 *          Size of buffer (excluding trailing null char).
 * @param	srcFileKind
 *          One of SRC_FILE_*
 * @return	True if successful.
 *****************************************************************************/
gboolean
HilitePerl( GtkWidget* pWidget,
            char*      pBuffer,
            int        bufferSize,
            int        srcFileKind )
{
   const char*	pc;
   const char*	pcAfter;
   const char*	pcTmp1;
   const char*	pcTmp2;
   const char*	pcPrevious;
   const char*	pcPrePastWhitespace;
   const char*	pCommentAfter;
   char*		pScratch;
   int			textWidgetSize;
   int          breakout = 0;

#define REMAINING_BUFFER_SIZE(p)	(bufferSize - (p - pBuffer))

  /*
   * Allocate mem for a temp string as large as module buffer.
   * Will be re-used many times by lower-level functions.
   */
   pScratch = g_malloc( bufferSize );

  /*
   * Parse buffer for strings to hilite.
   */
   pcPrevious = NULL;
   for ( pc = pBuffer; pc < &pBuffer[bufferSize]; )
   {
     /*
      * Break loop if pc is stuck (due to parser bug).
      */
      if ( pc == pcPrevious )
      {
         ++breakout;
         if ( breakout > 100 )
         {
            Warning( "Perl hilite code is stuck." );
            WarningMalfunction();
            break;
         }
      }
      pcPrevious = pc;

     /*
      * Note: Loop should be stopped using break, not return.
      */
g_return_val_if_fail( pc >= pBuffer  &&  pc < &pBuffer[bufferSize], FALSE );

     /*
      * Has end of buffer been reached?
      */
g_return_val_if_fail( pc != NULL, FALSE );	/* should not be NULL at this point */
      if ( pc >= &pc[bufferSize]  ||  *pc == '\0' )
         break;

     /*
      * Move past whitespace, but remember prior pointer.
      */
      pcPrePastWhitespace = pc;
      pc = (char*)PastWhitespace( pc, REMAINING_BUFFER_SIZE(pc) );

     /*
      * Does only whitespace remain?
      */
      if ( pc == NULL )
      {
         gtk_text_insert( GTK_TEXT(pWidget),
                          NULL,
                          NULL,
                          NULL,
                          pcPrePastWhitespace,
                          REMAINING_BUFFER_SIZE(pcPrePastWhitespace ) );
         break;
      }

     /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

     /*
      * Is the beginning of a comment being pointed to?
      */
      if ( pc[0] == '#' )
      {
        /*
         * Parse/insert comment.
         */
         pCommentAfter = \
         HiliteComment( pWidget,
                        pcPrePastWhitespace,
                        "\n",
                        REMAINING_BUFFER_SIZE(pcPrePastWhitespace) );

        /*
         * Was the end of the comment parsed?
         */
         if ( pCommentAfter )
         {
           /*
            * From after comment, iterate.
            */
            pc = pCommentAfter;
            continue;
         }
         else
         {
           /*
            * End of buffer must have been prematurely encountered.
            */
            break;
         }
      }

     /*
      * Is this a variable?
      */
      if ( IsPerlVarChar( *pc ) )
      {
         const char* pWord;
         int         sizeWord;

        /*
         * Parse what follows [$@%].
         */
         sizeWord = ParsePerlArg( pc, &pBuffer[bufferSize] );
         if ( sizeWord <= 0 )
         {
           /*
            * This is a Perl scalar/list/hash variable (not $_[n]).
            */
            sizeWord = ParseWordAlphanum( pc + 1,
                                          &pWord, /* !! don't use -- might become NULL !! */
                                          &pBuffer[bufferSize] - pc );
            if ( !sizeWord ) sizeWord = 1; /* assume a single-char variable (eg "$&") */
         }

        /*
         * Insert the parsed word along with preceding whitespace.
         */
         pcAfter = pc + 1 + sizeWord;

         gtk_text_insert( GTK_TEXT(pWidget),
                          pFontType,
                          &colorType,
                          NULL,
                          pcPrePastWhitespace,
                          pcAfter - pcPrePastWhitespace );

         pc = pcAfter;
         continue;
      }

     /*
      * Is this the beginning of a regex range (or array subscript)?
      * A regex range could contain a quote char which otherwise would trip HiliteQuotedString().
      */
      if ( pc[0] == '[' )
      {
         for ( pcAfter = pc; *pcAfter && *pcAfter != ']' && *pcAfter != '\n'; ++pcAfter ) { };
         if ( *pcAfter == ']' ) ++pcAfter;
         gtk_text_insert( GTK_TEXT(pWidget),
                          pFontSymbol,
                          &colorSymbol,
                          NULL,
                          pcPrePastWhitespace,
                          pcAfter - pcPrePastWhitespace );
         pc = pcAfter;
         continue;
      }

     /*
      * Is this a quoted/backticked string?
      */
      if ( *pc == '"' || *pc == '\'' || *pc == '`' )
      {
        /*
         * Does backslash precede it?
         */
         if ( pc == pBuffer || *(pc-1) != '\\' )
         {
            if ( HiliteQuotedString( *pc,
                                     pWidget,
                                     &colorString,
                                     NULL, /* default font */
                                     pcPrePastWhitespace,
                                     pc,
                                     &pBuffer[bufferSize],
                                     &pcAfter ) == FUNC_OK )
            {
               pc = pcAfter;
               continue;
            }
            else
            {
              /*
               * This can normally happen in Perl "here" files
               * which sometimes don't really contain Perl code.
               * Even README.pl files exist which are plain text.
               */
               goto InsertRemainder;
            }
         }
         else
         {
           /*
            * A backslash precedes it, just hilite the single quote char.
            */
            pcAfter = pc + 1;
            gtk_text_insert( GTK_TEXT(pWidget),
                             pFontSymbol,
                             &colorSymbol,
                             NULL,
                             pcPrePastWhitespace,
                             pcAfter - pcPrePastWhitespace );
            pc = pcAfter;
            continue;
         }
      }

     /*
      * Try to scoop contiguous symbol chars.
      */
      if ( ! isalpha( *pc ) )	/* perhaps non-alpha is a symbol char? */
      {
        /*
         * For each symbol char (excluding some that are specially hilited) after any whitespace...
         */
         for( pcAfter = pc;
              tablePerlSymbolMost[ (int)*pcAfter ]; /* null char isn't a symbol char */
              ++pcAfter )
         {
#ifdef HYPERSRC_PEDANTIC
g_return_val_if_fail( pcAfter < &pBuffer[bufferSize], FALSE );
#endif
           /*
            * Break if a dquote is found unless it is the first char.
            * If dquote is first, then the dquote was earlier determined
            * as not part of a dquoted string (eg it could be '"').
            */
            if ( *pcAfter == '"'  &&  pcAfter > pc ) break;

           /*
            * Break if a slash is found unless it is the first char.
            * If slash is first, then the slash was earlier determined
            * as not part of a comment.
            */
            if ( *pcAfter == '/'  &&  pcAfter > pc ) break;
         }

        /*
         * Were any symbol chars counted?
         */
         if ( pcAfter > pc )
         {
            gtk_text_insert( GTK_TEXT(pWidget),
                             pFontSymbol,
                             &colorSymbol,
                             NULL,
                             pcPrePastWhitespace,
                             pcAfter - pcPrePastWhitespace );
           /*
            * Move pointer, continue (stopping cases are checked above).
            */
            pc = pcAfter;
            continue;
         }

        /*
         * Fall thru.
         */
      }

     /*
      * Scoop all text until:
      * start of Perl comment, or symbol char (excluding dquote), or buffer end.
      */
      pcTmp1 = strstr( pc, "#" );
      pcTmp2 = FindPerlSymbolVarChar( pc, REMAINING_BUFFER_SIZE(pc), INCLUDE_VAR_CHARS );

      if ( pcTmp1 == NULL )					/* process lower ptr unless it is NULL */
         pcAfter = pcTmp2;
      else if ( pcTmp2 == NULL )
         pcAfter = pcTmp1;
      else
         pcAfter = MIN( pcTmp1, pcTmp2 );	/* neither was NULL */

      if ( pcAfter == NULL )				/* was both NULL (buffer end)? */
      {
        /*
         * Encountered buffer end.
         */
         pcAfter = &pBuffer[bufferSize];
      }

      HilitePerlString( pWidgetTextActive,
                        pcPrePastWhitespace,
                        pcAfter - pcPrePastWhitespace,
                        pcPrePastWhitespace > pBuffer ? *(pcPrePastWhitespace-1) : 0,
                        pScratch,
                        srcFileKind );

     /*
      * Move pointer, continue (stopping cases are checked above).
      */
      pc = pcAfter;
   }

  /*
   * There's a bug here if the size of the text widget
   * does not equal the size of the buffer.
   */
   textWidgetSize = gtk_text_get_length( GTK_TEXT(pWidgetTextActive) );
#ifdef HYPERSRC_DEBUG
   if ( textWidgetSize != bufferSize )
   {
      DebugMsg( "gtk_text_get_length() != bufferSize in module %s", ActiveModuleName() );
      DebugMsg( "gtk_text_get_length() = %d  bufferSize = %d", textWidgetSize, bufferSize );
   }
#endif

  /*
   * Dealloc, return.
   */
Return:
   g_free( pScratch );
   return TRUE;

InsertRemainder:
   gtk_text_insert( GTK_TEXT(pWidget),
                    NULL,
                    NULL,
                    NULL,
                    pcPrePastWhitespace,
                    (pBuffer + bufferSize) - pcPrePastWhitespace );
   goto Return;
}

/*****************************************************************************
 * Parse a string of code (text outside of comment enclosures).
 * Insert the string as hilited text into a text widget.
 * @param	pWidget
 *          Pointer to a text widget.
 * @param	pStr
 *			Pointer to a string of code (uncommented).
 * @param	strSize
 *          Size of string of code.
 * @param	precedingChar
 *          Pass zero if none (ie true beginning of buffer).
 * @param	pScratch
 *          Should point to a buffer as large as corresponding module buffer.
 * @param	srcFileKind
 *          One of SRC_FILE_*
 * @param	pFont[Keyword,Type,Symbol,..] (GLOBAL)
 *          Hilite font (can be NULL).
 * @param	color[Keyword,Type,Symbol,..] (GLOBAL)
 *          Hilite color.
 * @return	True if no errors.
 *****************************************************************************/
gboolean
HilitePerlString( GtkWidget*  pWidget,
                  const char* pStr,
                  int         strSize,
                  char        precedingChar,
                  char*       pScratch,
                  int         srcFileKind )
{
   const char*	pc = NULL;
   const char*	pWord;
   int          sizeWord;

   if ( strSize == 0 ) return TRUE;

g_return_val_if_fail( pStr && strSize > 0, FALSE );

   for ( pc = pStr;
         pc < &pStr[strSize];
         pc = (char*)PastWord( pWord, sizeWord ) )
   {
     /*
      * Parse next word.
      */
      sizeWord = ParseWordAlphanum( pc,
                                    &pWord,
                                    strSize - (pc - pStr) );

     /*
      * If no word was parsed (or word was too big to parse),
      * insert remainder into text widget, return.
      */
      if ( sizeWord <= 0  ||  pWord == NULL )	/* ParseWord() can return negative */
      {
         gtk_text_insert( GTK_TEXT(pWidget),
                          NULL,
                          NULL,
                          NULL,
                          pc,
                          strSize - (pc - pStr) );
         return TRUE;
      }

#ifdef HYPERSRC_PEDANTIC
g_return_val_if_fail( pWord >= pc  &&  pWord < &pStr[strSize], FALSE );
#endif

     /*
      * For comparing strings, extract word into a ASCIIZ string.
      */
      memcpy( pScratch, pWord, sizeWord );
      pScratch[sizeWord] = '\0';

     /*
      * Is this word a keyword?
      */
      if ( FindWord( keywords_perl, pScratch, sizeWord ) )
      {
        /*
         * Insert the parsed word along with preceding whitespace.
         */
         gtk_text_insert( GTK_TEXT(pWidget),
                          pFontKeyword,
                          &colorKeyword,
                          NULL,
                          pc,
                          (pWord - pc) + sizeWord );
         continue;
      }

     /*
      * Is this word a symbol char?
      */
      if ( FindWord( symbols_perl_other, pScratch, sizeWord ) )
      {
        /*
         * Insert the parsed word along with preceding whitespace.
         */
         gtk_text_insert( GTK_TEXT(pWidget),
                          pFontSymbol,
                          &colorSymbol,
                          NULL,
                          pc,
                          (pWord - pc) + sizeWord );
         continue;
      }

     /*
      * Is this word "goto"?
      */
      if ( strcmp( pScratch, "goto" ) == 0 )
      {
        /*
         * Insert the parsed word along with preceding whitespace.
         */
         gtk_text_insert( GTK_TEXT(pWidget),
                          pFontKeyword,
                          &colorGoto,
                          NULL,
                          pc,
                          (pWord - pc) + sizeWord );
         continue;
      }

     /*
      * Is this word a label?
      */
      if ( pWord[sizeWord] == ':' )
      {
         gtk_text_insert( GTK_TEXT(pWidget),
                          pFontKeyword,  /* no special label font */
                          &colorLabel,   /* use special label color */
                          NULL,
                          pc,
                          (pWord - pc) + sizeWord );
         continue;
      }

     /*
      * This word was not recognized.
      * Using normal font/color, insert the parsed word
      * along with preceding whitespace.
      */
      gtk_text_insert( GTK_TEXT(pWidget),
                       NULL,
                       NULL,
                       NULL,
                       pc,
                       (pWord - pc) + sizeWord );

     /*
      * Iterate...
      */
   }

   return FALSE; /* won't reach here but stifles compiler */
}

/*****************************************************************************
 * Parse a Perl arg: $_[n]
 * @param	pc
 *          Pointer that is possibly pointing DIRECTLY to a Perl arg.
 * @param	pBufEnd
 *          End of buffer.
 * @return	If pc is pointing directly to a Perl arg, returns its size,
 *          otherwise returns zero.
 *****************************************************************************/
int
ParsePerlArg( const char* pc,
              const char* pBufEnd )
{
  /*
   * Don't read past end of buffer.
   */
   if ( !pc  ||  (pc+4) > pBufEnd )
      return 0;

  /*
   * Compare each char in $_[n] except n.
   */
   if (    pc[0] == '$'
        && pc[1] == '_'
        && pc[2] == '['
   /* skip pc[3] */
        && pc[4] == ']' )
      return 5; /* sizeof $_[n] */
   else
      return 0;
}

/******************************************************************************
 * @return True if a char is one of [$@%].
 ******************************************************************************/
int
IsPerlVarChar( char c )
{
   return index( "$@%", c ) != NULL;
}

/******************************************************************************
 * @return True if a char is a Perl symbol char (excluding "$@%").
 ******************************************************************************/
int
IsPerlSymbolChar( char c )
{
   return ! IsPerlVarChar(c)  &&  IsSymbolChar(c);
}

/******************************************************************************
 * Find a Perl symbol char (or also [$@%]) in a string.
 ******************************************************************************/
const char*
FindPerlSymbolVarChar( const char* pStr,
                       int         len,
                       gboolean    includeVarChars )
{
   const char*	pc;

#ifndef HYPERSRC_SPEED
g_return_val_if_fail( pStr  &&  len > 0  &&  len < MAX_TEXT_BUFFER_LEN, NULL );
#endif

  /*
   * Until a symbol (or Perl var) char, or past string, or null char.
   */
   for ( pc = pStr;
         pc < &pStr[len]  &&  *pc;
         ++pc )
   {
      if (     IsPerlSymbolChar( *pc )
           || (includeVarChars  &&  IsPerlVarChar( *pc )) )
         return pc;
   }

   return NULL;
}
